==> util-5.0.0-sources.jar: net/sf/eBus/util/IndexPool.java <==
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
//
// This library is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
//
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA
// 02111-1307 USA
//
// The Initial Developer of the Original Code is Charles W. Rapp.
// Portions created by Charles W. Rapp are
// Copyright (C) 2001 - 2005, 2011, 2013 - 2014. Charles W. Rapp.
// All Rights Reserved.
//

package net.sf.eBus.util;

import java.io.Serializable;
import java.util.Formatter;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * {@code IndexPool} provides integer index value reuse. When
 * you create an {@code IndexPool}, you may specify the minimum
 * and maximum allowed values (inclusive). You retrieve the next
 * available value using {@link #nextIndex} and return values to
 * the pool with {@link #returnIndex}. When the pool is
 * exhausted, {@link #nextIndex} throws an
 * {@code IllegalStateException}. The application is responsible
 * for returning unused indices to the pool by calling
 * {@link #returnIndex(int)}. The application is also responsible
 * for making sure that {@link #returnIndex(int)} calls contain
 * unique indicies and that the same index does not appear in the
 * pool multiple times.
 * <p>
 * {@code IndexPool} uses both an {@link AtomicInteger} and a
 * {@link ConcurrentLinkedDeque} (as a stack) to track available
 * indices. When {@link #nextIndex()} is called, it first checks
 * if the index stack is empty. If it is empty, then
 * {@link AtomicInteger#getAndIncrement()} is called to get the
 * next unused index. If the returned value is &gt; than the
 * maximum allowed index, then {@link IllegalStateException}
 * exception is thrown. If the index stack is not empty, then
 * the an index is popped off the stack and returned. The idea
 * behind this technique is to continually use the same indices
 * in the hopes this improves processor cache hits.
 * </p>
 * <p>
 * {@code IndexPool} has no memory between application
 * executions. If you need to persist index values between
 * executions, use {@link net.sf.eBus.util.IndexCache}.
 * </p>
 * <p>
 * <strong>Note:</strong> {@code IndexPool} <em>is</em>
 * synchronized. It is safe for multiple threads to access the
 * same {@code IndexPoool} instance without synchronizing.
 * </p>
 *
 * @see net.sf.eBus.util.IndexCache
 *
 * @author <a href="mailto:rapp@acm.org">Charles Rapp</a>
 */

public final class IndexPool
    implements Serializable
{
//---------------------------------------------------------------
// Member methods.
//

    //-----------------------------------------------------------
    // Constructors.
    //

    /**
     * Creates a pool with the default size.
     * <p>
     * The default settings are:
     * <ul>
     *   <li>
     *     Minimum index is zero (this will be the
     *     first assigned value).
     *   </li>
     *   <li>
     *     Maximum index is {@code Integer.MAX_VALUE}.
     *   </li>
     * </ul>
     */
    public IndexPool()
    {
        this (0, Integer.MAX_VALUE);
    } // end of IndexPool()

    /**
     * Creates a pool with the specified minimum and maximum
     * indicies (inclusive).
     * @param minIndex the minimum index ({@link #nextIndex} will
     * return this value first.)
     * @param maxIndex the maximum index (inclusive).
     * @throws IllegalArgumentException
     * if:
     * <ul>
     *   <li>
     *     {@code minIndex} is &lt;= zero.
     *   </li>
     *   <li>
     *     {@code maxIndex} is &lt;= zero.
     *   </li>
     *   <li>
     *     {@code minIndex} is &gt; {@code maxIndex}.
     *   </li>
     * </ul>
     */
    public IndexPool(final int minIndex, final int maxIndex)
        throws IllegalArgumentException
    {
        if (minIndex < 0)
        {
            throw (
                new IllegalArgumentException(
                    "minIndex < 0 (" +
                    Integer.toString(minIndex) +
                    ")"));
        }
        else if (maxIndex < 0)
        {
            throw (
                new IllegalArgumentException(
                    "maxIndex < 0 (" +
                    Integer.toString(maxIndex) +
                    ")"));
        }
        else if (minIndex > maxIndex)
        {
            throw (
                new IllegalArgumentException(
                    "minIndex (" +
                    Integer.toString(minIndex) +
                    " > maxIndex (" +
                    Integer.toString(maxIndex) +
                    ")"));
        }

        _nextIndex = new AtomicInteger(minIndex);
        _minIndex = minIndex;
        _maxIndex = maxIndex;
        _pool = new ConcurrentLinkedDeque<>();
    } // end of IndexPool(int, int)

    //
    // end of Constructors.
    //-----------------------------------------------------------

    //-----------------------------------------------------------
    // Object Method Overrides.
    //

    /**
     * Returns a textual representation of this index pool.
     * @return a textual representation of this index pool.
     */
    @Override
    public String toString()
    {
        final Formatter retval = new Formatter();
        final Iterator<Integer> pit;
        String separator = "";

        retval.format("minimum index = %,d%n", _minIndex);
        retval.format("maximum index = %,d%n", _maxIndex);

        retval.format("cached indices = {");
        for (pit = _pool.iterator();
             pit.hasNext() == true;
             separator = ", ")
        {
            retval.format("%s%,d", separator, pit.next());
        }
        retval.format("}%nnext index = %,d", _nextIndex.get());

        return (retval.toString());
    } // end of toString()

    //
    // end of Object Method Overrides.
    //-----------------------------------------------------------

    //-----------------------------------------------------------
    // Get methods.
    //

    /**
     * Returns the pool's minimum index.
     * @return the pool's minimum index.
     */
    public int minIndex()
    {
        return (_minIndex);
    } // end of minIndex()

    /**
     * Returns the pool's maximum index.
     * @return the pool's maximum index.
     */
    public int maxIndex()
    {
        return (_maxIndex);
    } // end of maxIndex()

    //
    // end of Get methods.
    //-----------------------------------------------------------

    /**
     * Returns the next available index.
     * @return the next available index.
     * @exception IllegalStateException
     * if all indices are in use and the pool is exhausted. This
     * may be due to a failure to return unused indicies to the
     * pool.
     */
    public int nextIndex()
        throws IllegalStateException
    {
        Integer retval = _pool.pollFirst();

        if (retval == null)
        {
            // Generate the next value.
            retval = _nextIndex.getAndIncrement();

            // Is this pool exhausted?
            if (retval> _maxIndex)
            {
                // Yes, it is very tired and needs rest.
                throw (
                    new IllegalStateException(
                        "index pool is exhausted"));
            }
        }

        return (retval);
    } // end of nextIndex()

    /**
     * Puts an index back into the pool for reuse.
     * <p>
     * <strong>Note:</strong> this method does not prevent
     * returning the same index to the pool multiple times. It
     * is the caller's responsibility to make sure that an index
     * appears only once in the pool.
     * </p>
     * @param index Put this index back in the pool.
     * @exception IndexOutOfBoundsException
     * if {@code index} is less than the minimum index value
     * or greater than or equal to the maximum index value.
     */
    public void returnIndex(final int index)
        throws IndexOutOfBoundsException
    {
        if (index < _minIndex)
        {
            throw (
                new IndexOutOfBoundsException(
                    Integer.toString(index) +
                    " < minimum index" +
                    Integer.toString(_minIndex)));
        }
        else if (index > _maxIndex)
        {
            throw (
                new IndexOutOfBoundsException(
                    Integer.toString(index) +
                    " > maximum index" +
                    Integer.toString(_maxIndex)));
        }
        else
        {
            // Put this index into the pool.
            _pool.add(index);
        }

        return;
    } // end of returnIndex(int)

    /**
     * Resets the index pool to its initial state. All indices
     * are available for use. Any outstanding indices are
     * invalid.
     */
    public void reset()
    {
        _nextIndex.set(_minIndex);
        _pool.clear();

        return;
    } // end of reset()

//---------------------------------------------------------------
// Member data.
//

    /**
     * The next index value to be returned.
     */
    private AtomicInteger _nextIndex;

    /**
     * The minimum index value which may be used.
     */
    private final int _minIndex;

    /**
     * The maximum index value which may be used (inclusive).
     */
    private final int _maxIndex;

    /**
     * Store returned indices here as a stack. The most recently
     * returned index is the next available index.
     */
    private ConcurrentLinkedDeque<Integer> _pool;

    //-----------------------------------------------------------
    // Constants.
    //

    /**
     *  This is eBus version 2.1.0.
     */
    private static final long serialVersionUID = 0x020204L;
} // end of IndexPool

//
// CHANGE LOG
// $Log: IndexPool.java,v $
// Revision 1.6  2007/02/12 20:44:55  charlesr
// Corrected comments.
//
// Revision 1.5  2005/10/27 14:59:18  charlesr
// Added following methods:
// + minIndex: returns pool's minimum index.
// + maxIndex: returns pool's maximum index.
// + available: returns number of available indices.
// + reset: resets pool to initial state.
//
// Revision 1.4  2005/07/21 00:00:59  charlesr
// Moved to Java 5:
// + Using generics.
// + Declaring _pool as a SortedSet rather than a TreeSet.
//
// Revision 1.3  2005/03/17 22:00:08  charlesr
// Removed IllegalStateException from returnIndex().
//
// Revision 1.2  2005/03/17 19:51:52  charlesr
// Made this class serializable.
//
// Revision 1.1  2005/03/10 19:24:07  charlesr
// Correct error in nextIndex().
//
// Revision 1.0  2003/11/20 01:46:15  charlesr
// Initial revision
//

==> util-5.3.2-sources.jar: net/sf/eBus/util/IndexPool.java <==
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
//
// This library is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
//
// Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330,
// Boston, MA
// 02111-1307 USA
//
// The Initial Developer of the Original Code is Charles W. Rapp.
// Portions created by Charles W. Rapp are
// Copyright (C) 2001 - 2005, 2011, 2013 - 2014. Charles W. Rapp.
// All Rights Reserved.
//

package net.sf.eBus.util;

import java.io.Serializable;
import java.util.Formatter;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedDeque;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * {@code IndexPool} provides integer index value reuse. When
 * you create an {@code IndexPool}, you may specify the minimum
 * and maximum allowed values (inclusive). You retrieve the next
 * available value using {@link #nextIndex} and return values to
 * the pool with {@link #returnIndex}. When the pool is
 * exhausted, {@link #nextIndex} throws an
 * {@code IllegalStateException}. The application is responsible
 * for returning unused indices to the pool by calling
 * {@link #returnIndex(int)}. The application is also responsible
 * for making sure that {@link #returnIndex(int)} calls contain
 * unique indicies and that the same index does not appear in the
 * pool multiple times.
 * <p>
 * {@code IndexPool} uses both an {@link AtomicInteger} and a
 * {@link ConcurrentLinkedDeque} (as a stack) to track available
 * indices. When {@link #nextIndex()} is called, it first checks
 * if the index stack is empty. If it is empty, then
 * {@link AtomicInteger#getAndIncrement()} is called to get the
 * next unused index. If the returned value is &gt; than the
 * maximum allowed index, then {@link IllegalStateException}
 * exception is thrown. If the index stack is not empty, then
 * the an index is popped off the stack and returned. The idea
 * behind this technique is to continually use the same indices
 * in the hopes this improves processor cache hits.
 * </p>
 * <p>
 * {@code IndexPool} has no memory between application
 * executions. If you need to persist index values between
 * executions, use {@link net.sf.eBus.util.IndexCache}.
 * </p>
 * <p>
 * <strong>Note:</strong> {@code IndexPool} <em>is</em>
 * synchronized. It is safe for multiple threads to access the
 * same {@code IndexPoool} instance without synchronizing.
 * </p>
 *
 * @see net.sf.eBus.util.IndexCache
 *
 * @author <a href="mailto:rapp@acm.org">Charles Rapp</a>
 */

public final class IndexPool
    implements Serializable
{
//---------------------------------------------------------------
// Member data.
//

    //-----------------------------------------------------------
    // Constants.
    //

    /**
     *  This is eBus version 2.1.0.
     */
    private static final long serialVersionUID = 0x050200L;

    //-----------------------------------------------------------
    // Locals.
    //

    /**
     * The next index value to be returned.
     */
    private AtomicInteger mNextIndex;

    /**
     * The minimum index value which may be used.
     */
    private final int mMinIndex;

    /**
     * The maximum index value which may be used (inclusive).
     */
    private final int mMaxIndex;

    /**
     * Store returned indices here as a stack. The most recently
     * returned index is the next available index.
     */
    private ConcurrentLinkedDeque<Integer> mPool;

//---------------------------------------------------------------
// Member methods.
//

    //-----------------------------------------------------------
    // Constructors.
    //

    /**
     * Creates a pool with the default size.
     * <p>
     * The default settings are:
     * <ul>
     *   <li>
     *     Minimum index is zero (this will be the
     *     first assigned value).
     *   </li>
     *   <li>
     *     Maximum index is {@code Integer.MAX_VALUE}.
     *   </li>
     * </ul>
     */
    public IndexPool()
    {
        this (0, Integer.MAX_VALUE);
    } // end of IndexPool()

    /**
     * Creates a pool with the specified minimum and maximum
     * indicies (inclusive).
     * @param minIndex the minimum index ({@link #nextIndex} will
     * return this value first.)
     * @param maxIndex the maximum index (inclusive).
     * @throws IllegalArgumentException
     * if:
     * <ul>
     *   <li>
     *     {@code minIndex} is &lt;= zero.
     *   </li>
     *   <li>
     *     {@code maxIndex} is &lt;= zero.
     *   </li>
     *   <li>
     *     {@code minIndex} is &gt; {@code maxIndex}.
     *   </li>
     * </ul>
     */
    public IndexPool(final int minIndex, final int maxIndex)
        throws IllegalArgumentException
    {
        if (minIndex < 0)
        {
            throw (
                new IllegalArgumentException(
                    "minIndex < 0 (" +
                    Integer.toString(minIndex) +
                    ")"));
        }
        else if (maxIndex < 0)
        {
            throw (
                new IllegalArgumentException(
                    "maxIndex < 0 (" +
                    Integer.toString(maxIndex) +
                    ")"));
        }
        else if (minIndex > maxIndex)
        {
            throw (
                new IllegalArgumentException(
                    "minIndex (" +
                    Integer.toString(minIndex) +
                    " > maxIndex (" +
                    Integer.toString(maxIndex) +
                    ")"));
        }

        mNextIndex = new AtomicInteger(minIndex);
        mMinIndex = minIndex;
        mMaxIndex = maxIndex;
        mPool = new ConcurrentLinkedDeque<>();
    } // end of IndexPool(int, int)

    //
    // end of Constructors.
    //-----------------------------------------------------------

    //-----------------------------------------------------------
    // Object Method Overrides.
    //

    /**
     * Returns a textual representation of this index pool.
     * @return a textual representation of this index pool.
     */
    @Override
    public String toString()
    {
        final Formatter retval = new Formatter();
        final Iterator<Integer> pit;
        String separator = "";

        retval.format("minimum index = %,d%n", mMinIndex);
        retval.format("maximum index = %,d%n", mMaxIndex);

        retval.format("cached indices = {");
        for (pit = mPool.iterator();
             pit.hasNext() == true;
             separator = ", ")
        {
            retval.format("%s%,d", separator, pit.next());
        }
        retval.format("}%nnext index = %,d", mNextIndex.get());

        return (retval.toString());
    } // end of toString()

    //
    // end of Object Method Overrides.
    //-----------------------------------------------------------

    //-----------------------------------------------------------
    // Get methods.
    //

    /**
     * Returns the pool's minimum index.
     * @return the pool's minimum index.
     */
    public int minIndex()
    {
        return (mMinIndex);
    } // end of minIndex()

    /**
     * Returns the pool's maximum index.
     * @return the pool's maximum index.
     */
    public int maxIndex()
    {
        return (mMaxIndex);
    } // end of maxIndex()

    //
    // end of Get methods.
    //-----------------------------------------------------------

    /**
     * Returns the next available index.
     * @return the next available index.
     * @exception IllegalStateException
     * if all indices are in use and the pool is exhausted. This
     * may be due to a failure to return unused indicies to the
     * pool.
     */
    public int nextIndex()
        throws IllegalStateException
    {
        Integer retval = mPool.pollFirst();

        if (retval == null)
        {
            // Generate the next value.
            retval = mNextIndex.getAndIncrement();

            // Is this pool exhausted?
            if (retval> mMaxIndex)
            {
                // Yes, it is very tired and needs rest.
                throw (
                    new IllegalStateException(
                        "index pool is exhausted"));
            }
        }

        return (retval);
    } // end of nextIndex()

    /**
     * Puts an index back into the pool for reuse.
     * <p>
     * <strong>Note:</strong> this method does not prevent
     * returning the same index to the pool multiple times. It
     * is the caller's responsibility to make sure that an index
     * appears only once in the pool.
     * </p>
     * @param index Put this index back in the pool.
     * @exception IndexOutOfBoundsException
     * if {@code index} is less than the minimum index value
     * or greater than or equal to the maximum index value.
     */
    public void returnIndex(final int index)
        throws IndexOutOfBoundsException
    {
        if (index < mMinIndex)
        {
            throw (
                new IndexOutOfBoundsException(
                    Integer.toString(index) +
                    " < minimum index" +
                    Integer.toString(mMinIndex)));
        }
        else if (index > mMaxIndex)
        {
            throw (
                new IndexOutOfBoundsException(
                    Integer.toString(index) +
                    " > maximum index" +
                    Integer.toString(mMaxIndex)));
        }
        else
        {
            // Put this index into the pool.
            mPool.add(index);
        }

        return;
    } // end of returnIndex(int)

    /**
     * Resets the index pool to its initial state. All indices
     * are available for use. Any outstanding indices are
     * invalid.
     */
    public void reset()
    {
        mNextIndex.set(mMinIndex);
        mPool.clear();

        return;
    } // end of reset()
} // end of IndexPool

//
// CHANGE LOG
// $Log: IndexPool.java,v $
// Revision 1.6  2007/02/12 20:44:55  charlesr
// Corrected comments.
//
// Revision 1.5  2005/10/27 14:59:18  charlesr
// Added following methods:
// + minIndex: returns pool's minimum index.
// + maxIndex: returns pool's maximum index.
// + available: returns number of available indices.
// + reset: resets pool to initial state.
//
// Revision 1.4  2005/07/21 00:00:59  charlesr
// Moved to Java 5:
// + Using generics.
// + Declaring _pool as a SortedSet rather than a TreeSet.
//
// Revision 1.3  2005/03/17 22:00:08  charlesr
// Removed IllegalStateException from returnIndex().
//
// Revision 1.2  2005/03/17 19:51:52  charlesr
// Made this class serializable.
//
// Revision 1.1  2005/03/10 19:24:07  charlesr
// Correct error in nextIndex().
//
// Revision 1.0  2003/11/20 01:46:15  charlesr
// Initial revision
//
